// Model for distributed building firm
#define S 3
#define H 3

chan done = [0] of {bool};
chan start = [0] of {bool};

// State
// Record whether a specialist is working
bool working[S] = false;
// Record in which house the specialist is working currently (if it is working)
byte house[H] = 0;
byte manager_house = 0;

proctype specialist(byte id) {
  bool msg;
  do
    :: house[id] == H
        -> break;
    :: start ? msg
        -> printf("Specialist %d working on house %d\n", id, house[id]);
           working[id] = true;  // Specialist working
           working[id] = false; // Specialist done
           printf("Specialist %d done with house %d\n", id, house[id]);
           done ! true;         // Inform team manager that I am done
           house[id]++;        // Get ready to work in next house
  od;
}

inline notify_start_specialists() {
  printf("Start work on house %d\n",manager_house);
  bool msg;
  byte specialist_id = 0;
  do
    :: specialist_id == S
        -> break;
    :: else
        -> start ! true;
           specialist_id++;
  od;
}

inline wait_for_specialists() {
  printf("Waiting for specialist to finish house %d\n",manager_house);
  bool msg;
  byte specialists_ready = 0;
  do
    :: specialists_ready == S
        -> break;
    :: done ? msg
        -> specialists_ready++;
  od;
}

proctype manager() {
  do
    :: manager_house == H
        -> break;
    :: else
        -> notify_start_specialists();
           wait_for_specialists();
           manager_house++;
  od;
}

init {
  run specialist(0);
  run specialist(1);
  run specialist(2);
  run manager();
}

// The properties below assume 3 specialist and 3 houses

ltl eventually_specialist_work {
  <>(working[0]) &&
  <>(working[1]) &&
  <>(working[2])
}

ltl always_work_in_same_house {
  []( (working[0] && working[1] && working[2]) ->
      (house[0] == house[1] && house[1] == house[2]) )
}

ltl if_specialist_working_manager_waits {
  [] ( (working[0] -> manager_house == house[0]) ||
       (working[1] -> manager_house == house[1]) ||
       (working[2] -> manager_house == house[2]) )
}
