/**
  * *Acknowledgement*
  * Exercise from course TDA294 Formal Methods for Software Development at
  * Chalmers University of Technology.
  * Website: http://www.cse.chalmers.se/edu/year/2017/course/TDA294_Formal_Methods_for_Software_Development/Exercises/TemporalModelChecking/Exercises_TemporalModelChecking.html
  *
  */


/***
 * GLOBAL SETUP
 ***/

#define M 3 /* number of clients */
#define S 2 /* number of servers */
#define N 2 /* number of proxy sockets */

mtype {data0, data1, data2 , data3}; /* Possible server payloads */
mtype payload [4];                   /* Which server hosts which payload */

chan local_network = [0] of {   /* Client-Proxy communication */
  byte, /* client_id  */
  byte, /* server_id  */
  chan  /* reply chan */
};

chan global_network = [1] of {  /* Proxy-Server communication */
  byte, /* server_id  */
  chan  /* reply chan */
};

bool client_done [M] = false; // Temporal variable to determine when the client is done
mtype client_received[M]; // Temporal variable to store the data that has been received
byte client_asked[M]; // Temporal variable to determine which server has been picked

bool failed_server[S] = false; // Temporal variable to determine whether a server has crashed

/***
 * CLIENT
 ***/
proctype client(byte id) {
  //mtype data_received;              /* data received */
  chan callback = [0] of { mtype }; /* local callback channel */

  do
    :: client_done[id] = false;
       /* Randomly select server */
       byte server_asked;                /* ID of the server to be asked */
       select(server_asked : 0..(S-1));
       client_asked[id] = server_asked;

       /*Send request*/
       printf("Client%d requesting Server%d\n", id, server_asked);
       local_network ! id , server_asked, callback ;
       callback ? client_received[id];
       client_done[id] = true;

       printf("Client%d got the answer %e\n", id, client_received[id]);
  od;
}

/***
 * SERVER
 ***/
proctype server(byte  id) {
  chan reply;
end:
  do
    :: global_network ?? eval(id), reply;
         printf ("Server%d working\n", id );
         reply ! (payload[id]);

    //:: failed_server[id] = true; // Server fails
    //   printf ("Server%d stop working\n", id );
    //   false; /* Block forever */
  od;
}

/***
 * PROXY
 ***/
active proctype proxy() {

  byte client_id; /* ID of the client making request */
  byte server_id; /* ID of server requested by client */
  chan  client_ch; /* Callback channel for client */
  mtype data;      /* Payload received from server */

  chan socket   [N] = [0] of {mtype}; /* Socket receiving server response */
  bool available[N] = true;           /* Not waiting for server response? */
  chan replies  [N];                  /* Relating sockets to client callbacks */

  byte i; /* For looping over sockets */

end:
  do

  /* Receive a request from a client */
  :: local_network ? (client_id, server_id, client_ch) ->

     printf("Proxy a request from Client%d to Server%d\n", client_id, server_id);

     i = 0; /* Loop to find available socket */
     do
     :: i < N ->
          if
          :: available[i] ->
               available[i] = false;
               /* make a request */
               replies[i] = client_ch;
               printf ("Proxy asking Server%d \n", server_id);
               global_network ! server_id, socket[i];
               break;
          :: else ->
               i++;
          fi
     :: else -> /* Wait for a socket to become available */
          if    /* .. 'abusing' that we know number of sockets */
          :: socket[0] ? data ->
               replies  [0] ! data;
               available[0] = true;
          :: socket[1] ? data ->
               replies  [1] ! data;
               available[1] = true;
          fi;
          i = 0; /* Restart search */
     od;

  /* Receive a response from a server */
  :: socket[0] ? (data) ->
       replies  [0] ! data;
       available[0] = true;
  :: socket[1] ? (data) ->
       replies  [1] ! data;
       available[1] = true;

  od;

}

/***
 * INIT
 ***/
init {
  printf ("init\n");
  /* check params */
  S >= 2;
  N >= 2;

  /* Assign payloads */
  payload[0] = data0;
  payload[1] = data1;
  payload[2] = data2;
  payload[3] = data3;

  byte i = 0; /* initialize the servers */
  do
  :: i >= S -> break;
  :: else   -> run server(i); i++;
  od;

  i = 0;      /* initialize the clients */
  do
  :: i >= M -> break;
  :: else   -> run client (i); i++
  od;
}


ltl client_ends { (([] !failed_server[client_asked[0]]) -> <>client_done[0]) &&
                  (([] !failed_server[client_asked[1]]) -> <>client_done[1]) &&
                  (([] !failed_server[client_asked[2]]) -> <>client_done[2])}

ltl client_rec { [] ( (client_done[0] -> client_received[0] == payload[client_asked[0]]) &&
                      (client_done[1] -> client_received[1] == payload[client_asked[1]]) &&
                      (client_done[2] -> client_received[2] == payload[client_asked[2]]) )  }
